
/*
    nwfsinfo.c - Print the info strings of a server, maybe sometime more.
    Copyright (C) 1996 by Volker Lendecke
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    Revision history:

	0.00  1996			Volker Lendecke
		Initial revision.
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <ncp/ncplib.h>

#include <libintl.h>
#define _(X) gettext(X)

static char *progname;

static void
usage(void)
{
	fprintf(stderr, _("usage: %s [pattern]\n"), progname);
}

static void
help(void)
{
	printf(_("\n"
		 "usage: %s [options]\n"), progname);
	printf(_("\n"
	       "-h             Print this help text\n"
	       "-S server      Server name to be used\n"
	       "\n"
	       "-d             Print Description Strings\n"
	       "-t             Print File Server's time\n"
	       "-i             Print File Server Information\n"
	       "\n"));
}

static void
print_info(struct ncp_file_server_info *info)
{
	printf(_("\n"
	         "Fileservername    %-48.48s\n\n"), info->ServerName);
	printf(_("Version           %d.%d Revision %c\n"),
	       info->FileServiceVersion, info->FileServiceSubVersion,
	       info->Revision + 'A');
	printf(_("Max. Connections  %d\n"),
	       info->MaximumServiceConnections);
	printf(_("currently in use  %d\n"),
	       info->ConnectionsInUse);
	printf(_("peak connections  %d\n"),
	       info->MaxConnectionsEverUsed);
	printf(_("Max. Volumes      %d\n"),
	       info->NumberMountedVolumes);
	printf(_("SFTLevel          %d\n"),
	       info->SFTLevel);
	printf(_("TTSLevel          %d\n"),
	       info->TTSLevel);
	printf(_("Accountversion    %d\n"),
	       info->AccountVersion);
	printf(_("Queueversion      %d\n"),
	       info->QueueVersion);
	printf(_("Printversion      %d\n"),
	       info->PrintVersion);
	printf(_("Virt.Consolvers.  %d\n"),
	       info->VirtualConsoleVersion);
	printf(_("RestrictionLevel  %d\n"),
	       info->RestrictionLevel);
	printf("\n");
	return;
}

int
main(int argc, char **argv)
{
	struct ncp_conn *conn;
	int opt;
	long err;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	
	progname = argv[0];

	if ((conn = ncp_initialize(&argc, argv, 0, &err)) == NULL)
	{
		com_err(argv[0], err, _("when initializing"));
		return 1;
	}
	while ((opt = getopt(argc, argv, "h?dti")) != EOF)
	{
		switch (opt)
		{
		case 'h':
		case '?':
			help();
			break;
		case 'd':
			{
				char strings[512];
				char *s;

				if (ncp_get_file_server_description_strings(conn,
								      strings)
				    != 0)
				{
					perror(_("could not get strings"));
					ncp_close(conn);
					return 1;
				}
				s = strings;
				while (s < strings + 512)
				{
					if (strlen(s) == 0)
					{
						break;
					}
					puts(s);
					s += strlen(s) + 1;
				}
				break;
			}
		case 't':
			{
				time_t t;

				if (ncp_get_file_server_time(conn, &t) != 0)
				{
					perror(_("could not get server time"));
					ncp_close(conn);
					return 1;
				}
				{
					char text_server_time[200];
					struct tm* tm;
					
					tm = localtime(&t);
					strftime(text_server_time, sizeof(text_server_time), "%c", tm);
					printf("%s\n", text_server_time);
				}
				break;
			}
		case 'i':
			{
				struct ncp_file_server_info info;
				if (ncp_get_file_server_information(conn, &info) != 0)
				{
					perror(_("Could not get server information"));
					ncp_close(conn);
					return 1;
				}
				print_info(&info);
				break;
			}

		default:
			usage();
			goto finished;
		}
	}

      finished:
	ncp_close(conn);
	return 0;
}
